#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v2 Role action implementations"""

import logging

from keystoneauth1 import exceptions as ks_exc
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class AddRole(command.ShowOne):
    _description = _("Add role to project:user")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to add to <project>:<user> (name or ID)'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            required=True,
            help=_('Include <project> (name or ID)'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            required=True,
            help=_('Include <user> (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        role = utils.find_resource(identity_client.roles, parsed_args.role)
        project = utils.find_resource(
            identity_client.tenants,
            parsed_args.project,
        )
        user = utils.find_resource(identity_client.users, parsed_args.user)
        role = identity_client.roles.add_user_role(
            user.id,
            role.id,
            project.id,
        )

        info = {}
        info.update(role._info)
        return zip(*sorted(info.items()))


class CreateRole(command.ShowOne):
    _description = _("Create new role")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role_name',
            metavar='<name>',
            help=_('New role name'),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing role'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        try:
            role = identity_client.roles.create(parsed_args.role_name)
        except ks_exc.Conflict:
            if parsed_args.or_show:
                role = utils.find_resource(
                    identity_client.roles,
                    parsed_args.role_name,
                )
                LOG.info(_('Returning existing role %s'), role.name)
            else:
                raise

        info = {}
        info.update(role._info)
        return zip(*sorted(info.items()))


class DeleteRole(command.Command):
    _description = _("Delete role(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'roles',
            metavar='<role>',
            nargs="+",
            help=_('Role(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        errors = 0
        for role in parsed_args.roles:
            try:
                role_obj = utils.find_resource(
                    identity_client.roles,
                    role,
                )
                identity_client.roles.delete(role_obj.id)
            except Exception as e:
                errors += 1
                LOG.error(
                    _(
                        "Failed to delete role with "
                        "name or ID '%(role)s': %(e)s"
                    ),
                    {'role': role, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.roles)
            msg = _("%(errors)s of %(total)s roles failed to delete.") % {
                'errors': errors,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListRole(command.Lister):
    _description = _("List roles")

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        columns = ('ID', 'Name')
        data = identity_client.roles.list()

        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class RemoveRole(command.Command):
    _description = _("Remove role from project : user")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to remove (name or ID)'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            required=True,
            help=_('Include <project> (name or ID)'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            required=True,
            help=_('Include <user> (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        role = utils.find_resource(identity_client.roles, parsed_args.role)
        project = utils.find_resource(
            identity_client.tenants,
            parsed_args.project,
        )
        user = utils.find_resource(identity_client.users, parsed_args.user)
        identity_client.roles.remove_user_role(user.id, role.id, project.id)


class ShowRole(command.ShowOne):
    _description = _("Display role details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to display (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        role = utils.find_resource(identity_client.roles, parsed_args.role)

        info = {}
        info.update(role._info)
        return zip(*sorted(info.items()))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.network.v2 import router
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestRouter(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.projects_mock = self.identity_client.projects


class TestAddPortToRouter(TestRouter):
    '''Add port to Router'''

    _port = network_fakes.create_one_port()
    _router = network_fakes.create_one_router(attrs={'port': _port.id})

    def setUp(self):
        super().setUp()

        self.network_client.find_router.return_value = self._router
        self.network_client.find_port.return_value = self._port

        self.cmd = router.AddPortToRouter(self.app, None)

    def test_add_port_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_add_port_required_options(self):
        arglist = [
            self._router.id,
            self._router.port,
        ]
        verifylist = [
            ('router', self._router.id),
            ('port', self._router.port),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.add_interface_to_router.assert_called_with(
            self._router,
            **{
                'port_id': self._router.port,
            },
        )
        self.assertIsNone(result)


class TestAddSubnetToRouter(TestRouter):
    '''Add subnet to Router'''

    _subnet = network_fakes.FakeSubnet.create_one_subnet()
    _router = network_fakes.create_one_router(attrs={'subnet': _subnet.id})

    def setUp(self):
        super().setUp()

        self.network_client.find_router.return_value = self._router
        self.network_client.find_subnet.return_value = self._subnet

        self.cmd = router.AddSubnetToRouter(self.app, None)

    def test_add_subnet_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_add_subnet_required_options(self):
        arglist = [
            self._router.id,
            self._router.subnet,
        ]
        verifylist = [
            ('router', self._router.id),
            ('subnet', self._router.subnet),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.add_interface_to_router.assert_called_with(
            self._router, **{'subnet_id': self._router.subnet}
        )

        self.assertIsNone(result)


class TestCreateRouter(TestRouter):
    # The new router created.
    new_router = network_fakes.create_one_router()
    _extensions = {'fake': network_fakes.create_one_extension()}

    columns = (
        'admin_state_up',
        'availability_zone_hints',
        'availability_zones',
        'created_at',
        'description',
        'distributed',
        'enable_ndp_proxy',
        'external_gateway_info',
        'flavor_id',
        'ha',
        'id',
        'name',
        'project_id',
        'revision_number',
        'routes',
        'status',
        'tags',
        'updated_at',
    )
    data = (
        router.AdminStateColumn(new_router.is_admin_state_up),
        format_columns.ListColumn(new_router.availability_zone_hints),
        format_columns.ListColumn(new_router.availability_zones),
        new_router.created_at,
        new_router.description,
        new_router.is_distributed,
        new_router.enable_ndp_proxy,
        router.RouterInfoColumn(new_router.external_gateway_info),
        new_router.flavor_id,
        new_router.is_ha,
        new_router.id,
        new_router.name,
        new_router.project_id,
        new_router.revision_number,
        router.RoutesColumn(new_router.routes),
        new_router.status,
        format_columns.ListColumn(new_router.tags),
        new_router.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.network_client.create_router.return_value = self.new_router

        self.network_client.set_tags.return_value = None
        self.network_client.find_extension.side_effect = (
            lambda name, ignore_missing=True: self._extensions.get(name)
        )
        # Get the command object to test
        self.cmd = router.CreateRouter(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )
        self.assertFalse(self.network_client.set_tags.called)

    def test_create_default_options(self):
        arglist = [
            self.new_router.name,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
            }
        )
        self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_gateway(self):
        _network = network_fakes.create_one_network()
        _subnet = network_fakes.FakeSubnet.create_one_subnet()
        self.network_client.find_network.return_value = _network
        self.network_client.find_subnet.return_value = _subnet
        arglist = [
            self.new_router.name,
            '--external-gateway',
            _network.name,
            '--enable-snat',
            '--fixed-ip',
            'ip-address=2001:db8::1',
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('external_gateways', [_network.name]),
            ('enable_snat', True),
            ('fixed_ips', [{'ip-address': '2001:db8::1'}]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'external_gateway_info': {
                    'network_id': _network.id,
                    'enable_snat': True,
                    'external_fixed_ips': [{'ip_address': '2001:db8::1'}],
                },
            }
        )
        self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def _test_create_with_ha_options(self, option, ha):
        arglist = [
            option,
            self.new_router.name,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', ha),
            ('no_ha', not ha),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'ha': ha,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_ha_option(self):
        self._test_create_with_ha_options('--ha', True)

    def test_create_with_no_ha_option(self):
        self._test_create_with_ha_options('--no-ha', False)

    def _test_create_with_distributed_options(self, option, distributed):
        arglist = [
            option,
            self.new_router.name,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', distributed),
            ('centralized', not distributed),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'distributed': distributed,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_distributed_option(self):
        self._test_create_with_distributed_options('--distributed', True)

    def test_create_with_centralized_option(self):
        self._test_create_with_distributed_options('--centralized', False)

    def test_create_with_AZ_hints(self):
        arglist = [
            self.new_router.name,
            '--availability-zone-hint',
            'fake-az',
            '--availability-zone-hint',
            'fake-az2',
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('availability_zone_hints', ['fake-az', 'fake-az2']),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'availability_zone_hints': ['fake-az', 'fake-az2'],
            }
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def _test_create_with_tag(self, add_tags=True):
        arglist = [self.new_router.name]
        if add_tags:
            arglist += ['--tag', 'red', '--tag', 'blue']
        else:
            arglist += ['--no-tag']
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
        ]
        if add_tags:
            verifylist.append(('tags', ['red', 'blue']))
        else:
            verifylist.append(('no_tag', True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_router.assert_called_once_with(
            name=self.new_router.name, admin_state_up=True
        )
        if add_tags:
            self.network_client.set_tags.assert_called_once_with(
                self.new_router, tests_utils.CompareBySet(['red', 'blue'])
            )
        else:
            self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_tags(self):
        self._test_create_with_tag(add_tags=True)

    def test_create_with_no_tag(self):
        self._test_create_with_tag(add_tags=False)

    def test_create_with_flavor_id_id(self):
        _flavor = network_fakes.create_one_network_flavor()
        self.network_client.find_flavor.return_value = _flavor
        arglist = [
            self.new_router.name,
            '--flavor-id',
            _flavor.id,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('flavor_id', _flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'flavor_id': _flavor.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_flavor_id_name(self):
        _flavor = network_fakes.create_one_network_flavor()
        self.network_client.find_flavor.return_value = _flavor
        arglist = [self.new_router.name, '--flavor-id', _flavor.name]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('flavor_id', _flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'flavor_id': _flavor.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_flavor_id(self):
        _flavor = network_fakes.create_one_network_flavor()
        self.network_client.find_flavor.return_value = _flavor
        arglist = [
            self.new_router.name,
            '--flavor',
            _flavor.id,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('flavor', _flavor.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'flavor_id': _flavor.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_flavor_name(self):
        _flavor = network_fakes.create_one_network_flavor()
        self.network_client.find_flavor.return_value = _flavor
        arglist = [self.new_router.name, '--flavor', _flavor.name]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('flavor', _flavor.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                'flavor_id': _flavor.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_enable_default_route_bfd(self):
        self._extensions = {
            'external-gateway-multihoming': network_fakes.create_one_extension(
                attrs={'name': 'external-gateway-multihoming'}
            ),
        }

        arglist = [self.new_router.name, '--enable-default-route-bfd']
        verifylist = [
            ('name', self.new_router.name),
            ('enable_default_route_bfd', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            name=self.new_router.name,
            admin_state_up=True,
            enable_default_route_bfd=True,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_enable_default_route_bfd_no_extension(self):
        arglist = [self.new_router.name, '--enable-default-route-bfd']
        verifylist = [
            ('name', self.new_router.name),
            ('enable_default_route_bfd', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_create_with_enable_default_route_ecmp(self):
        self._extensions = {
            'external-gateway-multihoming': network_fakes.create_one_extension(
                attrs={'name': 'external-gateway-multihoming'}
            ),
        }

        arglist = [self.new_router.name, '--enable-default-route-ecmp']
        verifylist = [
            ('name', self.new_router.name),
            ('enable_default_route_ecmp', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.create_router.assert_called_once_with(
            name=self.new_router.name,
            admin_state_up=True,
            enable_default_route_ecmp=True,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_enable_default_route_ecmp_no_extension(self):
        arglist = [self.new_router.name, '--enable-default-route-ecmp']
        verifylist = [
            ('name', self.new_router.name),
            ('enable_default_route_ecmp', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_create_with_qos_policy(self):
        _network = network_fakes.create_one_network()
        self.network_client.find_network.return_value = _network
        _qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = _qos_policy

        arglist = [
            self.new_router.name,
            '--external-gateway',
            _network.id,
            '--qos-policy',
            _qos_policy.id,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('qos_policy', _qos_policy.id),
            ('external_gateways', [_network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        gw_info = {'network_id': _network.id, 'qos_policy_id': _qos_policy.id}
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self.new_router.name,
                **{'external_gateway_info': gw_info},
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_qos_policy_no_external_gateway(self):
        _qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = _qos_policy

        arglist = [
            self.new_router.name,
            '--qos-policy',
            _qos_policy.id,
        ]
        verifylist = [
            ('name', self.new_router.name),
            ('enable', True),
            ('distributed', False),
            ('ha', False),
            ('qos_policy', _qos_policy.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestDeleteRouter(TestRouter):
    # The routers to delete.
    _routers = network_fakes.create_routers(count=2)

    def setUp(self):
        super().setUp()

        self.network_client.delete_router.return_value = None

        self.network_client.find_router = network_fakes.get_routers(
            self._routers
        )

        # Get the command object to test
        self.cmd = router.DeleteRouter(self.app, None)

    def test_router_delete(self):
        arglist = [
            self._routers[0].name,
        ]
        verifylist = [
            ('router', [self._routers[0].name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.delete_router.assert_called_once_with(
            self._routers[0]
        )
        self.assertIsNone(result)

    def test_multi_routers_delete(self):
        arglist = []
        verifylist = []

        for r in self._routers:
            arglist.append(r.name)
        verifylist = [
            ('router', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for r in self._routers:
            calls.append(call(r))
        self.network_client.delete_router.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_routers_delete_with_exception(self):
        arglist = [
            self._routers[0].name,
            'unexist_router',
        ]
        verifylist = [
            ('router', [self._routers[0].name, 'unexist_router']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self._routers[0], exceptions.CommandError]
        self.network_client.find_router.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 routers failed to delete.', str(e))

        self.network_client.find_router.assert_any_call(
            self._routers[0].name, ignore_missing=False
        )
        self.network_client.find_router.assert_any_call(
            'unexist_router', ignore_missing=False
        )
        self.network_client.delete_router.assert_called_once_with(
            self._routers[0]
        )


class TestListRouter(TestRouter):
    # The routers going to be listed up.
    routers = network_fakes.create_routers(count=3)
    extensions = network_fakes.create_one_extension()

    columns = (
        'ID',
        'Name',
        'Status',
        'State',
        'Project',
        'Distributed',
        'HA',
    )
    columns_long = columns + (
        'Routes',
        'External gateway info',
        'Availability zones',
        'Tags',
    )
    columns_long_no_az = columns + (
        'Routes',
        'External gateway info',
        'Tags',
    )

    data = []
    for r in routers:
        data.append(
            (
                r.id,
                r.name,
                r.status,
                router.AdminStateColumn(r.is_admin_state_up),
                r.project_id,
                r.is_distributed,
                r.is_ha,
            )
        )

    router_agent_data = []
    for r in routers:
        router_agent_data.append(
            (
                r.id,
                r.name,
                r.external_gateway_info,
            )
        )

    agents_columns = (
        'ID',
        'Name',
        'External Gateway Info',
    )

    data_long = []
    for i in range(0, len(routers)):
        r = routers[i]
        data_long.append(
            data[i]
            + (
                router.RoutesColumn(r.routes),
                router.RouterInfoColumn(r.external_gateway_info),
                format_columns.ListColumn(r.availability_zones),
                format_columns.ListColumn(r.tags),
            )
        )
    data_long_no_az = []
    for i in range(0, len(routers)):
        r = routers[i]
        data_long_no_az.append(
            data[i]
            + (
                router.RoutesColumn(r.routes),
                router.RouterInfoColumn(r.external_gateway_info),
                format_columns.ListColumn(r.tags),
            )
        )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = router.ListRouter(self.app, None)

        self.network_client.agent_hosted_routers.return_value = self.routers

        self.network_client.routers.return_value = self.routers
        self.network_client.find_extension.return_value = self.extensions

        self.network_client.find_router.return_value = self.routers[0]

        self._testagent = network_fakes.create_one_network_agent()
        self.network_client.get_agent.return_value = self._testagent
        self.network_client.get_router.return_value = self.routers[0]

    def test_router_list_no_options(self):
        arglist = []
        verifylist = [
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_router_list_no_ha_no_distributed(self):
        _routers = network_fakes.create_routers(
            {'ha': None, 'distributed': None}, count=3
        )

        arglist = []
        verifylist = [
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(
            self.network_client, "routers", return_value=_routers
        ):
            columns, data = self.cmd.take_action(parsed_args)

        self.assertNotIn("is_distributed", columns)
        self.assertNotIn("is_ha", columns)

    def test_router_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with()
        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, list(data))

    def test_router_list_long_no_az(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # to mock, that no availability zone
        self.network_client.find_extension.return_value = None

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with()
        self.assertEqual(self.columns_long_no_az, columns)
        self.assertCountEqual(self.data_long_no_az, list(data))

    def test_list_name(self):
        test_name = "fakename"
        arglist = [
            '--name',
            test_name,
        ]
        verifylist = [
            ('long', False),
            ('name', test_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with(
            **{'name': test_name}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_router_list_enable(self):
        arglist = [
            '--enable',
        ]
        verifylist = [
            ('long', False),
            ('enable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with(
            **{'admin_state_up': True, 'is_admin_state_up': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_router_list_disable(self):
        arglist = [
            '--disable',
        ]
        verifylist = [('long', False), ('disable', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with(
            **{'admin_state_up': False, 'is_admin_state_up': False}
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_router_list_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.routers.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_router_list_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
            '--project-domain',
            project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
            ('project_domain', project.domain_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.routers.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_router_list_agents_no_args(self):
        arglist = [
            '--agents',
        ]
        verifylist = []

        # Missing required router ID should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_router_list_agents(self):
        arglist = [
            '--agent',
            self._testagent.id,
        ]
        verifylist = [
            ('agent', self._testagent.id),
        ]

        attrs = {
            self._testagent.id,
        }

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.agent_hosted_routers(*attrs)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_list_with_tag_options(self):
        arglist = [
            '--tags',
            'red,blue',
            '--any-tags',
            'red,green',
            '--not-tags',
            'orange,yellow',
            '--not-any-tags',
            'black,white',
        ]
        verifylist = [
            ('tags', ['red', 'blue']),
            ('any_tags', ['red', 'green']),
            ('not_tags', ['orange', 'yellow']),
            ('not_any_tags', ['black', 'white']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers.assert_called_once_with(
            **{
                'tags': 'red,blue',
                'any_tags': 'red,green',
                'not_tags': 'orange,yellow',
                'not_any_tags': 'black,white',
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))


class TestRemovePortFromRouter(TestRouter):
    '''Remove port from a Router'''

    _port = network_fakes.create_one_port()
    _router = network_fakes.create_one_router(attrs={'port': _port.id})

    def setUp(self):
        super().setUp()

        self.network_client.find_router.return_value = self._router
        self.network_client.find_port.return_value = self._port

        self.cmd = router.RemovePortFromRouter(self.app, None)

    def test_remove_port_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_remove_port_required_options(self):
        arglist = [
            self._router.id,
            self._router.port,
        ]
        verifylist = [
            ('router', self._router.id),
            ('port', self._router.port),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.remove_interface_from_router.assert_called_with(
            self._router, **{'port_id': self._router.port}
        )
        self.assertIsNone(result)


class TestRemoveSubnetFromRouter(TestRouter):
    '''Remove subnet from Router'''

    _subnet = network_fakes.FakeSubnet.create_one_subnet()
    _router = network_fakes.create_one_router(attrs={'subnet': _subnet.id})

    def setUp(self):
        super().setUp()

        self.network_client.find_router.return_value = self._router
        self.network_client.find_subnet.return_value = self._subnet

        self.cmd = router.RemoveSubnetFromRouter(self.app, None)

    def test_remove_subnet_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_remove_subnet_required_options(self):
        arglist = [
            self._router.id,
            self._router.subnet,
        ]
        verifylist = [
            ('subnet', self._router.subnet),
            ('router', self._router.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_interface_from_router.assert_called_with(
            self._router, **{'subnet_id': self._router.subnet}
        )
        self.assertIsNone(result)


class TestAddExtraRoutesToRouter(TestRouter):
    _router = network_fakes.create_one_router()

    def setUp(self):
        super().setUp()
        self.network_client.add_extra_routes_to_router.return_value = (
            self._router
        )

        self.cmd = router.AddExtraRoutesToRouter(self.app, None)
        self.network_client.find_router.return_value = self._router

    def test_add_no_extra_route(self):
        arglist = [
            self._router.id,
        ]
        verifylist = [
            ('router', self._router.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.add_extra_routes_to_router.assert_called_with(
            self._router, body={'router': {'routes': []}}
        )
        self.assertEqual(2, len(result))

    def test_add_one_extra_route(self):
        arglist = [
            self._router.id,
            '--route',
            'destination=dst1,gateway=gw1',
        ]
        verifylist = [
            ('router', self._router.id),
            ('routes', [{'destination': 'dst1', 'gateway': 'gw1'}]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.add_extra_routes_to_router.assert_called_with(
            self._router,
            body={
                'router': {
                    'routes': [
                        {'destination': 'dst1', 'nexthop': 'gw1'},
                    ]
                }
            },
        )
        self.assertEqual(2, len(result))

    def test_add_multiple_extra_routes(self):
        arglist = [
            self._router.id,
            '--route',
            'destination=dst1,gateway=gw1',
            '--route',
            'destination=dst2,gateway=gw2',
        ]
        verifylist = [
            ('router', self._router.id),
            (
                'routes',
                [
                    {'destination': 'dst1', 'gateway': 'gw1'},
                    {'destination': 'dst2', 'gateway': 'gw2'},
                ],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.add_extra_routes_to_router.assert_called_with(
            self._router,
            body={
                'router': {
                    'routes': [
                        {'destination': 'dst1', 'nexthop': 'gw1'},
                        {'destination': 'dst2', 'nexthop': 'gw2'},
                    ]
                }
            },
        )
        self.assertEqual(2, len(result))


class TestRemoveExtraRoutesFromRouter(TestRouter):
    _router = network_fakes.create_one_router()

    def setUp(self):
        super().setUp()
        self.network_client.remove_extra_routes_from_router.return_value = (
            self._router
        )

        self.cmd = router.RemoveExtraRoutesFromRouter(self.app, None)
        self.network_client.find_router.return_value = self._router

    def test_remove_no_extra_route(self):
        arglist = [
            self._router.id,
        ]
        verifylist = [
            ('router', self._router.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.remove_extra_routes_from_router.assert_called_with(
            self._router, body={'router': {'routes': []}}
        )
        self.assertEqual(2, len(result))

    def test_remove_one_extra_route(self):
        arglist = [
            self._router.id,
            '--route',
            'destination=dst1,gateway=gw1',
        ]
        verifylist = [
            ('router', self._router.id),
            ('routes', [{'destination': 'dst1', 'gateway': 'gw1'}]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.remove_extra_routes_from_router.assert_called_with(
            self._router,
            body={
                'router': {
                    'routes': [
                        {'destination': 'dst1', 'nexthop': 'gw1'},
                    ]
                }
            },
        )
        self.assertEqual(2, len(result))

    def test_remove_multiple_extra_routes(self):
        arglist = [
            self._router.id,
            '--route',
            'destination=dst1,gateway=gw1',
            '--route',
            'destination=dst2,gateway=gw2',
        ]
        verifylist = [
            ('router', self._router.id),
            (
                'routes',
                [
                    {'destination': 'dst1', 'gateway': 'gw1'},
                    {'destination': 'dst2', 'gateway': 'gw2'},
                ],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.remove_extra_routes_from_router.assert_called_with(
            self._router,
            body={
                'router': {
                    'routes': [
                        {'destination': 'dst1', 'nexthop': 'gw1'},
                        {'destination': 'dst2', 'nexthop': 'gw2'},
                    ]
                }
            },
        )
        self.assertEqual(2, len(result))


class TestSetRouter(TestRouter):
    # The router to set.
    _default_route = {'destination': '10.20.20.0/24', 'nexthop': '10.20.30.1'}
    _network = network_fakes.create_one_network()
    _subnet = network_fakes.FakeSubnet.create_one_subnet(
        attrs={'network_id': _network.id}
    )
    _router = network_fakes.create_one_router(
        attrs={'routes': [_default_route], 'tags': ['green', 'red']}
    )
    _extensions = {'fake': network_fakes.create_one_extension()}

    def setUp(self):
        super().setUp()
        self.network_client.update_router.return_value = None
        self.network_client.set_tags.return_value = None
        self.network_client.find_router.return_value = self._router
        self.network_client.find_network.return_value = self._network

        self.network_client.find_subnet.return_value = self._subnet
        self.network_client.find_extension.side_effect = (
            lambda name, ignore_missing=True: self._extensions.get(name)
        )
        # Get the command object to test
        self.cmd = router.SetRouter(self.app, None)

    def test_set_this(self):
        arglist = [
            self._router.name,
            '--enable',
            '--distributed',
            '--name',
            'noob',
            '--no-ha',
            '--description',
            'router',
        ]
        verifylist = [
            ('router', self._router.name),
            ('enable', True),
            ('distributed', True),
            ('name', 'noob'),
            ('description', 'router'),
            ('no_ha', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'admin_state_up': True,
            'distributed': True,
            'name': 'noob',
            'ha': False,
            'description': 'router',
        }
        self.network_client.update_router.assert_called_once_with(
            self._router, **attrs
        )
        self.assertIsNone(result)

    def test_set_that(self):
        arglist = [
            self._router.name,
            '--disable',
            '--centralized',
            '--ha',
        ]
        verifylist = [
            ('router', self._router.name),
            ('disable', True),
            ('centralized', True),
            ('ha', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'admin_state_up': False,
            'distributed': False,
            'ha': True,
        }
        self.network_client.update_router.assert_called_once_with(
            self._router, **attrs
        )
        self.assertIsNone(result)

    def test_set_distributed_centralized(self):
        arglist = [
            self._router.name,
            '--distributed',
            '--centralized',
        ]
        verifylist = [
            ('router', self._router.name),
            ('distributed', True),
            ('distributed', False),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_route(self):
        arglist = [
            self._router.name,
            '--route',
            'destination=10.20.30.0/24,gateway=10.20.30.1',
        ]
        verifylist = [
            ('router', self._router.name),
            (
                'routes',
                [{'destination': '10.20.30.0/24', 'gateway': '10.20.30.1'}],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        routes = [{'destination': '10.20.30.0/24', 'nexthop': '10.20.30.1'}]
        attrs = {'routes': routes + self._router.routes}
        self.network_client.update_router.assert_called_once_with(
            self._router, **attrs
        )
        self.assertIsNone(result)

    def test_set_no_route(self):
        arglist = [
            self._router.name,
            '--no-route',
        ]
        verifylist = [
            ('router', self._router.name),
            ('no_route', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'routes': [],
        }
        self.network_client.update_router.assert_called_once_with(
            self._router, **attrs
        )
        self.assertIsNone(result)

    def test_set_route_overwrite_route(self):
        _testrouter = network_fakes.create_one_router(
            {'routes': [{"destination": "10.0.0.2", "nexthop": "1.1.1.1"}]}
        )
        self.network_client.find_router.return_value = _testrouter
        arglist = [
            _testrouter.name,
            '--route',
            'destination=10.20.30.0/24,gateway=10.20.30.1',
            '--no-route',
        ]
        verifylist = [
            ('router', _testrouter.name),
            (
                'routes',
                [{'destination': '10.20.30.0/24', 'gateway': '10.20.30.1'}],
            ),
            ('no_route', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'routes': [
                {'destination': '10.20.30.0/24', 'nexthop': '10.20.30.1'}
            ]
        }
        self.network_client.update_router.assert_called_once_with(
            _testrouter, **attrs
        )
        self.assertIsNone(result)

    def test_set_nothing(self):
        arglist = [
            self._router.name,
        ]
        verifylist = [
            ('router', self._router.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_router.called)
        self.assertFalse(self.network_client.set_tags.called)
        self.assertIsNone(result)

    def test_wrong_gateway_params(self):
        arglist = [
            "--fixed-ip",
            "subnet='abc'",
            self._router.id,
        ]
        verifylist = [
            ('fixed_ips', [{'subnet': "'abc'"}]),
            ('router', self._router.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_set_gateway_network_only(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            self._router.id,
        ]
        verifylist = [
            ('external_gateways', [self._network.id]),
            ('router', self._router.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.update_router.assert_called_with(
            self._router,
            **{'external_gateway_info': {'network_id': self._network.id}},
        )
        self.assertIsNone(result)

    def test_set_gateway_options_subnet_only(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            "--fixed-ip",
            "subnet='abc'",
            self._router.id,
            '--enable-snat',
        ]
        verifylist = [
            ('router', self._router.id),
            ('external_gateways', [self._network.id]),
            ('fixed_ips', [{'subnet': "'abc'"}]),
            ('enable_snat', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.update_router.assert_called_with(
            self._router,
            **{
                'external_gateway_info': {
                    'network_id': self._network.id,
                    'external_fixed_ips': [
                        {
                            'subnet_id': self._subnet.id,
                        }
                    ],
                    'enable_snat': True,
                }
            },
        )
        self.assertIsNone(result)

    def test_set_gateway_option_ipaddress_only(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            "--fixed-ip",
            "ip-address=10.0.1.1",
            self._router.id,
            '--enable-snat',
        ]
        verifylist = [
            ('router', self._router.id),
            ('external_gateways', [self._network.id]),
            ('fixed_ips', [{'ip-address': "10.0.1.1"}]),
            ('enable_snat', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.update_router.assert_called_with(
            self._router,
            **{
                'external_gateway_info': {
                    'network_id': self._network.id,
                    'external_fixed_ips': [
                        {
                            'ip_address': "10.0.1.1",
                        }
                    ],
                    'enable_snat': True,
                }
            },
        )
        self.assertIsNone(result)

    def test_set_gateway_options_subnet_ipaddress(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            "--fixed-ip",
            "subnet='abc',ip-address=10.0.1.1",
            self._router.id,
            '--enable-snat',
        ]
        verifylist = [
            ('router', self._router.id),
            ('external_gateways', [self._network.id]),
            ('fixed_ips', [{'subnet': "'abc'", 'ip-address': "10.0.1.1"}]),
            ('enable_snat', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.update_router.assert_called_with(
            self._router,
            **{
                'external_gateway_info': {
                    'network_id': self._network.id,
                    'external_fixed_ips': [
                        {
                            'subnet_id': self._subnet.id,
                            'ip_address': "10.0.1.1",
                        }
                    ],
                    'enable_snat': True,
                }
            },
        )
        self.assertIsNone(result)

    def _test_set_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['red', 'blue', 'green']
        else:
            arglist = ['--no-tag']
            verifylist = [('no_tag', True)]
            expected_args = []
        arglist.append(self._router.name)
        verifylist.append(('router', self._router.name))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_router.called)
        self.network_client.set_tags.assert_called_once_with(
            self._router, tests_utils.CompareBySet(expected_args)
        )
        self.assertIsNone(result)

    def test_set_with_tags(self):
        self._test_set_tags(with_tags=True)

    def test_set_with_no_tag(self):
        self._test_set_tags(with_tags=False)

    def test_set_gateway_ip_qos(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy

        arglist = [
            "--external-gateway",
            self._network.id,
            "--qos-policy",
            qos_policy.id,
            self._router.id,
        ]
        verifylist = [
            ('router', self._router.id),
            ('external_gateways', [self._network.id]),
            ('qos_policy', qos_policy.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.update_router.assert_called_with(
            self._router,
            **{
                'external_gateway_info': {
                    'network_id': self._network.id,
                    'qos_policy_id': qos_policy.id,
                }
            },
        )
        self.assertIsNone(result)

    def test_unset_gateway_ip_qos(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            "--no-qos-policy",
            self._router.id,
        ]
        verifylist = [
            ('router', self._router.id),
            ('external_gateways', [self._network.id]),
            ('no_qos_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.update_router.assert_called_with(
            self._router,
            **{
                'external_gateway_info': {
                    'network_id': self._network.id,
                    'qos_policy_id': None,
                }
            },
        )
        self.assertIsNone(result)

    def test_set_unset_gateway_ip_qos(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy

        arglist = [
            "--external-gateway",
            self._network.id,
            "--qos-policy",
            qos_policy.id,
            "--no-qos-policy",
            self._router.id,
        ]
        verifylist = [
            ('router', self._router.id),
            ('external_gateways', [self._network.id]),
            ('qos_policy', qos_policy.id),
            ('no_qos_policy', True),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_gateway_ip_qos_no_gateway(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy

        router = network_fakes.create_one_router()
        self.network_client.find_router.return_value = router
        arglist = [
            "--qos-policy",
            qos_policy.id,
            router.id,
        ]
        verifylist = [
            ('router', router.id),
            ('qos_policy', qos_policy.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_unset_gateway_ip_qos_no_gateway(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy

        router = network_fakes.create_one_router()
        self.network_client.find_router.return_value = router
        arglist = [
            "--no-qos-policy",
            router.id,
        ]
        verifylist = [
            ('router', router.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShowRouter(TestRouter):
    # The router to set.
    _router = network_fakes.create_one_router()
    _port = network_fakes.create_one_port(
        {'device_owner': 'network:router_interface', 'device_id': _router.id}
    )
    setattr(
        _router,
        'interfaces_info',
        [
            {
                'port_id': _port.id,
                'ip_address': _port.fixed_ips[0]['ip_address'],
                'subnet_id': _port.fixed_ips[0]['subnet_id'],
            }
        ],
    )

    columns = (
        'admin_state_up',
        'availability_zone_hints',
        'availability_zones',
        'created_at',
        'description',
        'distributed',
        'enable_ndp_proxy',
        'external_gateway_info',
        'flavor_id',
        'ha',
        'id',
        'interfaces_info',
        'name',
        'project_id',
        'revision_number',
        'routes',
        'status',
        'tags',
        'updated_at',
    )
    data = (
        router.AdminStateColumn(_router.is_admin_state_up),
        format_columns.ListColumn(_router.availability_zone_hints),
        format_columns.ListColumn(_router.availability_zones),
        _router.created_at,
        _router.description,
        _router.is_distributed,
        _router.enable_ndp_proxy,
        router.RouterInfoColumn(_router.external_gateway_info),
        _router.flavor_id,
        _router.is_ha,
        _router.id,
        router.RouterInfoColumn(_router.interfaces_info),
        _router.name,
        _router.project_id,
        _router.revision_number,
        router.RoutesColumn(_router.routes),
        _router.status,
        format_columns.ListColumn(_router.tags),
        _router.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.network_client.find_router.return_value = self._router
        self.network_client.ports.return_value = [self._port]

        # Get the command object to test
        self.cmd = router.ShowRouter(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self._router.name,
        ]
        verifylist = [
            ('router', self._router.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_router.assert_called_once_with(
            self._router.name, ignore_missing=False
        )
        self.network_client.ports.assert_called_with(
            **{'device_id': self._router.id}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_show_no_ha_no_distributed(self):
        _router = network_fakes.create_one_router(
            {'ha': None, 'distributed': None}
        )

        arglist = [
            _router.name,
        ]
        verifylist = [
            ('router', _router.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(
            self.network_client, "find_router", return_value=_router
        ):
            columns, data = self.cmd.take_action(parsed_args)

        self.assertNotIn("is_distributed", columns)
        self.assertNotIn("is_ha", columns)

    def test_show_no_extra_route_extension(self):
        _router = network_fakes.create_one_router({'routes': None})

        arglist = [
            _router.name,
        ]
        verifylist = [
            ('router', _router.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(
            self.network_client, "find_router", return_value=_router
        ):
            columns, data = self.cmd.take_action(parsed_args)

        self.assertIn("routes", columns)
        self.assertIsNone(list(data)[columns.index('routes')].human_readable())


class TestUnsetRouter(TestRouter):
    def setUp(self):
        super().setUp()
        self.fake_network = network_fakes.create_one_network()
        self.fake_qos_policy = network_fakes.create_one_qos_policy()
        self._testrouter = network_fakes.create_one_router(
            {
                'routes': [
                    {
                        "destination": "192.168.101.1/24",
                        "nexthop": "172.24.4.3",
                    },
                    {
                        "destination": "192.168.101.2/24",
                        "nexthop": "172.24.4.3",
                    },
                ],
                'tags': ['green', 'red'],
                'external_gateway_info': {
                    'network_id': self.fake_network.id,
                    'qos_policy_id': self.fake_qos_policy.id,
                },
            }
        )
        self.fake_subnet = network_fakes.FakeSubnet.create_one_subnet()
        self.network_client.find_router.return_value = self._testrouter

        self.network_client.update_router.return_value = None
        self.network_client.set_tags.return_value = None
        self._extensions = {'fake': network_fakes.create_one_extension()}
        self.network_client.find_extension.side_effect = (
            lambda name, ignore_missing=True: self._extensions.get(name)
        )
        self.network_client.remove_external_gateways.return_value = None

        # Get the command object to test
        self.cmd = router.UnsetRouter(self.app, None)

    def test_unset_router_params(self):
        arglist = [
            '--route',
            'destination=192.168.101.1/24,gateway=172.24.4.3',
            self._testrouter.name,
        ]
        verifylist = [
            (
                'routes',
                [{"destination": "192.168.101.1/24", "gateway": "172.24.4.3"}],
            ),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'routes': [
                {"destination": "192.168.101.2/24", "nexthop": "172.24.4.3"}
            ],
        }
        self.network_client.update_router.assert_called_once_with(
            self._testrouter, **attrs
        )
        self.assertIsNone(result)

    def test_unset_router_wrong_routes(self):
        arglist = [
            '--route',
            'destination=192.168.101.1/24,gateway=172.24.4.2',
            self._testrouter.name,
        ]
        verifylist = [
            (
                'routes',
                [{"destination": "192.168.101.1/24", "gateway": "172.24.4.2"}],
            ),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_unset_router_external_gateway(self):
        arglist = [
            '--external-gateway',
            self._testrouter.name,
        ]
        verifylist = [('external_gateways', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {'external_gateway_info': {}}
        self.network_client.update_router.assert_called_once_with(
            self._testrouter, **attrs
        )
        self.assertIsNone(result)

    def test_unset_router_external_gateway_multiple_supported(self):
        # Add the relevant extension in order to test the alternate behavior.
        self._extensions = {
            'external-gateway-multihoming': network_fakes.create_one_extension(
                attrs={'name': 'external-gateway-multihoming'}
            )
        }
        arglist = [
            '--external-gateway',
            self._testrouter.name,
        ]
        verifylist = [('external_gateways', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        # The removal of all gateways should be requested using the multiple
        # gateways API.
        self.network_client.remove_external_gateways.assert_called_once_with(
            self._testrouter, body={'router': {'external_gateways': {}}}
        )
        # The compatibility API will also be called in order to potentially
        # unset other parameters along with external_gateway_info which
        # should already be empty at that point anyway.
        self.network_client.update_router.assert_called_once_with(
            self._testrouter, **{'external_gateway_info': {}}
        )
        self.assertIsNone(result)

    def _test_unset_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['green']
        else:
            arglist = ['--all-tag']
            verifylist = [('all_tag', True)]
            expected_args = []
        arglist.append(self._testrouter.name)
        verifylist.append(('router', self._testrouter.name))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_router.called)
        self.network_client.set_tags.assert_called_once_with(
            self._testrouter, tests_utils.CompareBySet(expected_args)
        )
        self.assertIsNone(result)

    def test_unset_with_tags(self):
        self._test_unset_tags(with_tags=True)

    def test_unset_with_all_tag(self):
        self._test_unset_tags(with_tags=False)

    def test_unset_router_qos_policy(self):
        arglist = [
            '--qos-policy',
            self._testrouter.name,
        ]
        verifylist = [('qos_policy', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'external_gateway_info': {
                "network_id": self.fake_network.id,
                "qos_policy_id": None,
            }
        }
        self.network_client.update_router.assert_called_once_with(
            self._testrouter, **attrs
        )
        self.assertIsNone(result)

    def test_unset_gateway_ip_qos_no_network(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy

        router = network_fakes.create_one_router()
        self.network_client.find_router.return_value = router
        arglist = [
            "--qos-policy",
            router.id,
        ]
        verifylist = [
            ('router', router.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_unset_gateway_ip_qos_no_qos(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy

        router = network_fakes.create_one_router(
            {"external_gateway_info": {"network_id": "fake-id"}}
        )
        self.network_client.find_router.return_value = router
        arglist = [
            "--qos-policy",
            router.id,
        ]
        verifylist = [
            ('router', router.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestGatewayOps(TestRouter):
    def setUp(self):
        super().setUp()
        self._networks = []
        self._network = network_fakes.create_one_network()
        self._networks.append(self._network)

        self._router = network_fakes.create_one_router(
            {
                'external_gateway_info': {
                    'network_id': self._network.id,
                },
            }
        )
        self._subnet = network_fakes.FakeSubnet.create_one_subnet(
            attrs={'network_id': self._network.id}
        )
        self._extensions = {
            'external-gateway-multihoming': network_fakes.create_one_extension(
                attrs={'name': 'external-gateway-multihoming'}
            )
        }
        self.network_client.find_extension.side_effect = (
            lambda name, ignore_missing=True: self._extensions.get(name)
        )
        self.network_client.find_router.return_value = self._router

        def _find_network(name_or_id, ignore_missing):
            for network in self._networks:
                if name_or_id in (network.id, network.name):
                    return network
            if ignore_missing:
                return None
            raise Exception('Test resource not found')

        self.network_client.find_network.side_effect = _find_network

        self.network_client.find_subnet.return_value = self._subnet
        self.network_client.add_external_gateways.return_value = None

        self.network_client.remove_external_gateways.return_value = None


class TestCreateMultipleGateways(TestGatewayOps):
    _columns = (
        'admin_state_up',
        'availability_zone_hints',
        'availability_zones',
        'created_at',
        'description',
        'distributed',
        'enable_ndp_proxy',
        'external_gateway_info',
        'flavor_id',
        'ha',
        'id',
        'name',
        'project_id',
        'revision_number',
        'routes',
        'status',
        'tags',
        'updated_at',
    )

    def setUp(self):
        super().setUp()
        self._second_network = network_fakes.create_one_network()
        self._networks.append(self._second_network)

        self.network_client.create_router.return_value = self._router

        self.network_client.update_router.return_value = None
        self.network_client.update_external_gateways.return_value = None

        self._data = (
            router.AdminStateColumn(self._router.is_admin_state_up),
            format_columns.ListColumn(self._router.availability_zone_hints),
            format_columns.ListColumn(self._router.availability_zones),
            self._router.created_at,
            self._router.description,
            self._router.is_distributed,
            self._router.enable_ndp_proxy,
            router.RouterInfoColumn(self._router.external_gateway_info),
            self._router.flavor_id,
            self._router.is_ha,
            self._router.id,
            self._router.name,
            self._router.project_id,
            self._router.revision_number,
            router.RoutesColumn(self._router.routes),
            self._router.status,
            format_columns.ListColumn(self._router.tags),
            self._router.updated_at,
        )
        self.cmd = router.CreateRouter(self.app, None)

    def test_create_one_gateway(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            self._router.name,
        ]
        verifylist = [
            ('name', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.update_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                        }
                    ]
                }
            },
        )
        self.assertEqual(self._columns, columns)
        self.assertCountEqual(self._data, data)

    def test_create_multiple_gateways(self):
        arglist = [
            self._router.name,
            "--external-gateway",
            self._network.id,
            "--external-gateway",
            self._network.id,
            "--external-gateway",
            self._second_network.id,
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.1',
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.2',
        ]
        verifylist = [
            ('name', self._router.name),
            (
                'external_gateways',
                [self._network.id, self._network.id, self._second_network.id],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # The router will not have a gateway after the create call, but it
        # will be added after the update call.
        self.network_client.create_router.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self._router.name,
            }
        )
        self.network_client.update_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.1',
                                }
                            ],
                        },
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.2',
                                }
                            ],
                        },
                        {
                            'network_id': self._second_network.id,
                        },
                    ]
                }
            },
        )
        self.assertEqual(self._columns, columns)
        self.assertCountEqual(self._data, data)


class TestUpdateMultipleGateways(TestGatewayOps):
    def setUp(self):
        super().setUp()
        self._second_network = network_fakes.create_one_network()
        self._networks.append(self._second_network)

        self.network_client.update_router.return_value = None
        self.network_client.update_external_gateways.return_value = None

        self.cmd = router.SetRouter(self.app, None)

    def test_update_one_gateway(self):
        arglist = [
            "--external-gateway",
            self._network.id,
            "--no-qos-policy",
            self._router.name,
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
            ('no_qos_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.update_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {'network_id': self._network.id, 'qos_policy_id': None}
                    ]
                }
            },
        )
        self.assertIsNone(result)

    def test_update_multiple_gateways(self):
        arglist = [
            self._router.name,
            "--external-gateway",
            self._network.id,
            "--external-gateway",
            self._network.id,
            "--external-gateway",
            self._second_network.id,
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.1',
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.2',
            "--no-qos-policy",
        ]
        verifylist = [
            ('router', self._router.name),
            (
                'external_gateways',
                [self._network.id, self._network.id, self._second_network.id],
            ),
            ('no_qos_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.update_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.1',
                                }
                            ],
                            'qos_policy_id': None,
                        },
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.2',
                                }
                            ],
                            'qos_policy_id': None,
                        },
                        {
                            'network_id': self._second_network.id,
                            'qos_policy_id': None,
                        },
                    ]
                }
            },
        )
        self.assertIsNone(result)


class TestAddGatewayRouter(TestGatewayOps):
    def setUp(self):
        super().setUp()
        # Get the command object to test
        self.cmd = router.AddGatewayToRouter(self.app, None)

        self.network_client.add_external_gateways.return_value = self._router

    def test_add_gateway_network_only(self):
        arglist = [
            self._router.name,
            self._network.id,
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.add_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [{'network_id': self._network.id}]
                }
            },
        )
        self.assertEqual(result[1][result[0].index('id')], self._router.id)

    def test_add_gateway_network_fixed_ip(self):
        arglist = [
            self._router.name,
            self._network.id,
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.1',
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.add_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.1',
                                }
                            ],
                        }
                    ]
                }
            },
        )
        self.assertEqual(result[1][result[0].index('id')], self._router.id)

    def test_add_gateway_network_multiple_fixed_ips(self):
        arglist = [
            self._router.name,
            self._network.id,
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.1',
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.2',
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
            (
                'fixed_ips',
                [
                    {'ip-address': '10.0.1.1', 'subnet': self._subnet.id},
                    {'ip-address': '10.0.1.2', 'subnet': self._subnet.id},
                ],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.add_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.1',
                                },
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.2',
                                },
                            ],
                        }
                    ]
                }
            },
        )
        self.assertEqual(result[1][result[0].index('id')], self._router.id)

    def test_add_gateway_network_only_no_extension(self):
        self._extensions = {}
        arglist = [
            self._router.name,
            self._network.id,
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestRemoveGatewayRouter(TestGatewayOps):
    def setUp(self):
        super().setUp()
        # Get the command object to test
        self.cmd = router.RemoveGatewayFromRouter(self.app, None)

        self.network_client.remove_external_gateways.return_value = (
            self._router
        )

    def test_remove_gateway_network_only(self):
        arglist = [
            self._router.name,
            self._network.id,
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [{'network_id': self._network.id}]
                }
            },
        )
        self.assertEqual(result[1][result[0].index('id')], self._router.id)

    def test_remove_gateway_network_fixed_ip(self):
        arglist = [
            self._router.name,
            self._network.id,
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.1',
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.1',
                                }
                            ],
                        }
                    ]
                }
            },
        )
        self.assertEqual(result[1][result[0].index('id')], self._router.id)

    def test_remove_gateway_network_multiple_fixed_ips(self):
        arglist = [
            self._router.name,
            self._network.id,
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.1',
            '--fixed-ip',
            f'subnet={self._subnet.id},ip-address=10.0.1.2',
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
            (
                'fixed_ips',
                [
                    {'ip-address': '10.0.1.1', 'subnet': self._subnet.id},
                    {'ip-address': '10.0.1.2', 'subnet': self._subnet.id},
                ],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_external_gateways.assert_called_with(
            self._router,
            body={
                'router': {
                    'external_gateways': [
                        {
                            'network_id': self._network.id,
                            'external_fixed_ips': [
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.1',
                                },
                                {
                                    'subnet_id': self._subnet.id,
                                    'ip_address': '10.0.1.2',
                                },
                            ],
                        }
                    ]
                }
            },
        )
        self.assertEqual(result[1][result[0].index('id')], self._router.id)

    def test_remove_gateway_network_only_no_extension(self):
        self._extensions = {}
        arglist = [
            self._router.name,
            self._network.id,
        ]
        verifylist = [
            ('router', self._router.name),
            ('external_gateways', [self._network.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.tests.functional.identity.v3 import common


class RegionTests(common.IdentityTests):
    def test_region_create(self):
        self._create_dummy_region()

    def test_region_create_with_parent_region(self):
        parent_region_id = self._create_dummy_region()
        self._create_dummy_region(parent_region=parent_region_id)

    def test_region_delete(self):
        region_id = self._create_dummy_region(add_clean_up=False)
        raw_output = self.openstack(f'region delete {region_id}')
        self.assertEqual(0, len(raw_output))

    def test_region_multi_delete(self):
        region_1 = self._create_dummy_region(add_clean_up=False)
        region_2 = self._create_dummy_region(add_clean_up=False)
        raw_output = self.openstack(f'region delete {region_1} {region_2}')
        self.assertEqual(0, len(raw_output))

    def test_region_list(self):
        raw_output = self.openstack('region list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.REGION_LIST_HEADERS)

    def test_region_set(self):
        # prepare region with parent-region
        parent_region_id = self._create_dummy_region()
        new_parent_region_id = self._create_dummy_region()
        region_id = self._create_dummy_region(parent_region_id)
        # check region details
        raw_output = self.openstack(f'region show {region_id}')
        region = self.parse_show_as_object(raw_output)
        self.assertEqual(parent_region_id, region['parent_region'])
        self.assertEqual(region_id, region['region'])
        # update parent-region
        raw_output = self.openstack(
            f'region set --parent-region {new_parent_region_id} {region_id}'
        )
        self.assertEqual(0, len(raw_output))
        # check updated region details
        raw_output = self.openstack(f'region show {region_id}')
        region = self.parse_show_as_object(raw_output)
        self.assertEqual(new_parent_region_id, region['parent_region'])
        self.assertEqual(region_id, region['region'])

    def test_region_show(self):
        region_id = self._create_dummy_region()
        raw_output = self.openstack(f'region show {region_id}')
        region = self.parse_show_as_object(raw_output)
        self.assertEqual(region_id, region['region'])
        self.assertEqual('None', region['parent_region'])



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Consumer action implementations"""

import logging

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreateConsumer(command.ShowOne):
    _description = _("Create new consumer")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New consumer description'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        consumer = identity_client.oauth1.consumers.create(
            parsed_args.description
        )
        consumer._info.pop('links', None)
        return zip(*sorted(consumer._info.items()))


class DeleteConsumer(command.Command):
    _description = _("Delete consumer(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'consumer',
            metavar='<consumer>',
            nargs='+',
            help=_('Consumer(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        result = 0
        for i in parsed_args.consumer:
            try:
                consumer = utils.find_resource(
                    identity_client.oauth1.consumers, i
                )
                identity_client.oauth1.consumers.delete(consumer.id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete consumer with name or "
                        "ID '%(consumer)s': %(e)s"
                    ),
                    {'consumer': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.consumer)
            msg = _("%(result)s of %(total)s consumers failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListConsumer(command.Lister):
    _description = _("List consumers")

    def take_action(self, parsed_args):
        columns = ('ID', 'Description')
        data = self.app.client_manager.identity.oauth1.consumers.list()
        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetConsumer(command.Command):
    _description = _("Set consumer properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'consumer',
            metavar='<consumer>',
            help=_('Consumer to modify'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New consumer description'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        consumer = utils.find_resource(
            identity_client.oauth1.consumers, parsed_args.consumer
        )
        kwargs = {}
        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        consumer = identity_client.oauth1.consumers.update(
            consumer.id, **kwargs
        )


class ShowConsumer(command.ShowOne):
    _description = _("Display consumer details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'consumer',
            metavar='<consumer>',
            help=_('Consumer to display'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        consumer = utils.find_resource(
            identity_client.oauth1.consumers, parsed_args.consumer
        )

        consumer._info.pop('links', None)
        return zip(*sorted(consumer._info.items()))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstackclient.image.v2 import metadef_resource_type_association
from openstackclient.tests.unit.image.v2 import fakes as resource_type_fakes


class TestMetadefResourceTypeAssociationCreate(
    resource_type_fakes.TestImagev2
):
    resource_type_association = (
        resource_type_fakes.create_one_resource_type_association()
    )

    columns = (
        'created_at',
        'id',
        'name',
        'prefix',
        'properties_target',
        'updated_at',
    )

    data = (
        resource_type_association.created_at,
        resource_type_association.id,
        resource_type_association.name,
        resource_type_association.prefix,
        resource_type_association.properties_target,
        resource_type_association.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.image_client.create_metadef_resource_type_association.return_value = self.resource_type_association
        self.cmd = metadef_resource_type_association.CreateMetadefResourceTypeAssociation(
            self.app, None
        )

    def test_resource_type_association_create(self):
        arglist = [
            self.resource_type_association.namespace_name,
            self.resource_type_association.name,
        ]

        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestMetadefResourceTypeAssociationDelete(
    resource_type_fakes.TestImagev2
):
    resource_type_association = (
        resource_type_fakes.create_one_resource_type_association()
    )

    def setUp(self):
        super().setUp()

        self.image_client.delete_metadef_resource_type_association.return_value = self.resource_type_association
        self.cmd = metadef_resource_type_association.DeleteMetadefResourceTypeAssociation(
            self.app, None
        )

    def test_resource_type_association_delete(self):
        arglist = [
            self.resource_type_association.namespace_name,
            self.resource_type_association.name,
        ]

        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)


class TestMetadefResourceTypeAssociationList(resource_type_fakes.TestImagev2):
    resource_type_associations = (
        resource_type_fakes.create_resource_type_associations()
    )

    columns = ['Name']

    datalist = [
        (resource_type_association.name,)
        for resource_type_association in resource_type_associations
    ]

    def setUp(self):
        super().setUp()

        self.image_client.metadef_resource_type_associations.side_effect = [
            self.resource_type_associations,
            [],
        ]

        self.cmd = metadef_resource_type_association.ListMetadefResourceTypeAssociations(
            self.app, None
        )

    def test_resource_type_association_list(self):
        arglist = [
            self.resource_type_associations[0].namespace_name,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class ServiceTests(common.IdentityTests):
    def test_service_create(self):
        self._create_dummy_service()

    def test_service_delete(self):
        service_name = self._create_dummy_service(add_clean_up=False)
        raw_output = self.openstack(f'service delete {service_name}')
        self.assertEqual(0, len(raw_output))

    def test_service_multi_delete(self):
        service_1 = self._create_dummy_service(add_clean_up=False)
        service_2 = self._create_dummy_service(add_clean_up=False)
        raw_output = self.openstack(f'service delete {service_1} {service_2}')
        self.assertEqual(0, len(raw_output))

    def test_service_list(self):
        self._create_dummy_service()
        raw_output = self.openstack('service list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)

    def test_service_set(self):
        service_name = self._create_dummy_service()
        # set service
        new_service_name = data_utils.rand_name('NewTestService')
        new_service_description = data_utils.rand_name('description')
        new_service_type = data_utils.rand_name('NewTestType')
        raw_output = self.openstack(
            'service set '
            f'--type {new_service_type} '
            f'--name {new_service_name} '
            f'--description {new_service_description} '
            '--disable '
            f'{service_name}'
        )
        self.assertEqual(0, len(raw_output))
        # get service details
        raw_output = self.openstack(f'service show {new_service_name}')
        # assert service details
        service = self.parse_show_as_object(raw_output)
        self.assertEqual(new_service_type, service['type'])
        self.assertEqual(new_service_name, service['name'])
        self.assertEqual(new_service_description, service['description'])

    def test_service_show(self):
        service_name = self._create_dummy_service()
        raw_output = self.openstack(f'service show {service_name}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.SERVICE_FIELDS)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Host action implementations"""

from openstack import exceptions as sdk_exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class ListHost(command.Lister):
    _description = _("DEPRECATED: List hosts")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--zone",
            metavar="<zone>",
            help=_("Only return hosts in the availability zone"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        self.log.warning(
            "API has been deprecated; "
            "consider using 'hypervisor list' instead."
        )

        # doing this since openstacksdk has decided not to support this
        # deprecated command
        response = compute_client.get('/os-hosts', microversion='2.1')
        sdk_exceptions.raise_from_response(response)
        hosts = response.json().get('hosts')
        if parsed_args.zone is not None:
            hosts = [h for h in hosts if h['zone'] == parsed_args.zone]

        columns = ("Host Name", "Service", "Zone")
        return columns, (utils.get_dict_properties(s, columns) for s in hosts)


class SetHost(command.Command):
    _description = _("DEPRECATED: Set host properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "host", metavar="<host>", help=_("Host to modify (name only)")
        )
        status = parser.add_mutually_exclusive_group()
        status.add_argument(
            '--enable', action='store_true', help=_("Enable the host")
        )
        status.add_argument(
            '--disable', action='store_true', help=_("Disable the host")
        )
        maintenance = parser.add_mutually_exclusive_group()
        maintenance.add_argument(
            '--enable-maintenance',
            action='store_true',
            help=_("Enable maintenance mode for the host"),
        )
        maintenance.add_argument(
            '--disable-maintenance',
            action='store_true',
            help=_("Disable maintenance mode for the host"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        self.log.warning(
            "API has been deprecated; "
            "consider using 'compute service set' instead."
        )

        data = {}
        if parsed_args.enable:
            data['status'] = 'enable'
        if parsed_args.disable:
            data['status'] = 'disable'
        if parsed_args.enable_maintenance:
            data['maintenance_mode'] = 'enable'
        if parsed_args.disable_maintenance:
            data['maintenance_mode'] = 'disable'

        if not data:
            # don't bother calling if nothing given
            return

        # doing this since openstacksdk has decided not to support this
        # deprecated command
        response = compute_client.put(
            f'/os-hosts/{parsed_args.host}', json=data, microversion='2.1'
        )
        sdk_exceptions.raise_from_response(response)


class ShowHost(command.Lister):
    _description = _("DEPRECATED: Display host details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("host", metavar="<host>", help=_("Name of host"))
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        self.log.warning(
            "API has been deprecated; "
            "consider using 'hypervisor show' instead."
        )

        # doing this since openstacksdk has decided not to support this
        # deprecated command
        response = compute_client.get(
            f'/os-hosts/{parsed_args.host}', microversion='2.1'
        )
        sdk_exceptions.raise_from_response(response)
        resources = response.json().get('host')

        data = []
        if resources is not None:
            for resource in resources:
                data.append(resource['resource'])

        columns = ("Host", "Project", "CPU", "Memory MB", "Disk GB")

        return columns, (utils.get_dict_properties(s, columns) for s in data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class RouterTests(common.NetworkTagTests):
    """Functional tests for router"""

    base_command = 'router'

    def test_router_create_and_delete(self):
        """Test create options, delete multiple"""
        name1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'router create ' + name1,
            parse_output=True,
        )
        self.assertEqual(
            name1,
            cmd_output["name"],
        )
        cmd_output = self.openstack(
            'router create ' + name2,
            parse_output=True,
        )
        self.assertEqual(
            name2,
            cmd_output["name"],
        )

        del_output = self.openstack('router delete ' + name1 + ' ' + name2)
        self.assertOutput('', del_output)

    def test_router_create_with_external_gateway(self):
        network_name = uuid.uuid4().hex
        subnet_name = uuid.uuid4().hex
        qos_policy = uuid.uuid4().hex
        router_name = uuid.uuid4().hex

        cmd_net = self.openstack(
            f'network create --external {network_name}', parse_output=True
        )
        self.addCleanup(self.openstack, f'network delete {network_name}')
        network_id = cmd_net['id']

        self.openstack(
            f'subnet create {subnet_name} '
            f'--network {network_name} --subnet-range 10.0.0.0/24'
        )

        cmd_qos = self.openstack(
            f'network qos policy create {qos_policy}', parse_output=True
        )
        self.addCleanup(
            self.openstack, f'network qos policy delete {qos_policy}'
        )
        qos_id = cmd_qos['id']

        self.openstack(
            f'router create --external-gateway {network_name} '
            f'--qos-policy {qos_policy} {router_name}'
        )
        self.addCleanup(self.openstack, f'router delete {router_name}')

        cmd_output = self.openstack(
            f'router show {router_name}', parse_output=True
        )
        gw_info = cmd_output['external_gateway_info']
        self.assertEqual(network_id, gw_info['network_id'])
        self.assertEqual(qos_id, gw_info['qos_policy_id'])

    def test_router_list(self):
        """Test create, list filter"""
        # Get project IDs
        cmd_output = self.openstack('token issue', parse_output=True)
        auth_project_id = cmd_output['project_id']

        cmd_output = self.openstack('project list', parse_output=True)
        admin_project_id = None
        demo_project_id = None
        for p in cmd_output:
            if p['Name'] == 'admin':
                admin_project_id = p['ID']
            if p['Name'] == 'demo':
                demo_project_id = p['ID']

        # Verify assumptions:
        # * admin and demo projects are present
        # * demo and admin are distinct projects
        # * tests run as admin
        self.assertIsNotNone(admin_project_id)
        self.assertIsNotNone(demo_project_id)
        self.assertNotEqual(admin_project_id, demo_project_id)
        self.assertEqual(admin_project_id, auth_project_id)

        # type narrow
        assert admin_project_id is not None
        assert demo_project_id is not None

        name1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'router create ' + '--disable ' + name1,
            parse_output=True,
        )

        self.addCleanup(self.openstack, 'router delete ' + name1)
        self.assertEqual(
            name1,
            cmd_output["name"],
        )
        self.assertEqual(
            False,
            cmd_output["admin_state_up"],
        )
        self.assertEqual(
            admin_project_id,
            cmd_output["project_id"],
        )
        cmd_output = self.openstack(
            'router create ' + '--project ' + demo_project_id + ' ' + name2,
            parse_output=True,
        )

        self.addCleanup(self.openstack, 'router delete ' + name2)
        self.assertEqual(
            name2,
            cmd_output["name"],
        )
        self.assertEqual(
            True,
            cmd_output["admin_state_up"],
        )
        self.assertEqual(
            demo_project_id,
            cmd_output["project_id"],
        )

        # Test list --project
        cmd_output = self.openstack(
            'router list ' + '--project ' + demo_project_id,
            parse_output=True,
        )
        names = [x["Name"] for x in cmd_output]
        self.assertNotIn(name1, names)
        self.assertIn(name2, names)

        # Test list --disable
        cmd_output = self.openstack(
            'router list ' + '--disable ',
            parse_output=True,
        )
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertNotIn(name2, names)

        # Test list --name
        cmd_output = self.openstack(
            'router list ' + '--name ' + name1,
            parse_output=True,
        )
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertNotIn(name2, names)

        # Test list --long
        cmd_output = self.openstack(
            'router list ' + '--long ',
            parse_output=True,
        )
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertIn(name2, names)

    def test_router_list_l3_agent(self):
        """Tests create router, add l3 agent, list, delete"""

        if not self.is_extension_enabled("l3_agent_scheduler"):
            self.skipTest("No l3_agent_scheduler extension present")

        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'router create ' + name,
            parse_output=True,
        )

        self.addCleanup(self.openstack, 'router delete ' + name)
        # Get router ID
        router_id = cmd_output['id']
        # Get l3 agent id
        cmd_output = self.openstack(
            'network agent list --agent-type l3',
            parse_output=True,
        )

        # Check at least one L3 agent is included in the response.
        self.assertTrue(cmd_output)
        agent_id = cmd_output[0]['ID']

        # Add router to agent
        self.openstack(
            'network agent add router --l3 ' + agent_id + ' ' + router_id
        )

        cmd_output = self.openstack(
            'router list --agent ' + agent_id,
            parse_output=True,
        )
        router_ids = [x['ID'] for x in cmd_output]
        self.assertIn(router_id, router_ids)

        # Remove router from agent
        self.openstack(
            'network agent remove router --l3 ' + agent_id + ' ' + router_id
        )
        cmd_output = self.openstack(
            'router list --agent ' + agent_id,
            parse_output=True,
        )
        router_ids = [x['ID'] for x in cmd_output]
        self.assertNotIn(router_id, router_ids)

    def test_router_set_show_unset(self):
        """Tests create router, set, unset, show"""

        name = uuid.uuid4().hex
        new_name = name + "_"
        cmd_output = self.openstack(
            'router create ' + '--description aaaa ' + name,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'router delete ' + new_name)
        self.assertEqual(
            name,
            cmd_output["name"],
        )
        self.assertEqual(
            'aaaa',
            cmd_output["description"],
        )

        # Test set --disable
        cmd_output = self.openstack(
            'router set '
            + '--name '
            + new_name
            + ' --description bbbb '
            + '--disable '
            + name
        )
        self.assertOutput('', cmd_output)

        cmd_output = self.openstack(
            'router show ' + new_name,
            parse_output=True,
        )
        self.assertEqual(
            new_name,
            cmd_output["name"],
        )
        self.assertEqual(
            'bbbb',
            cmd_output["description"],
        )
        self.assertEqual(
            False,
            cmd_output["admin_state_up"],
        )

        # Test set --ha --distributed
        self._test_set_router_distributed(new_name)

        # Test unset
        cmd_output = self.openstack(
            'router unset ' + '--external-gateway ' + new_name
        )
        cmd_output = self.openstack(
            'router show ' + new_name,
            parse_output=True,
        )
        self.assertIsNone(cmd_output["external_gateway_info"])

    def _test_set_router_distributed(self, router_name):
        if not self.is_extension_enabled("dvr"):
            return

        cmd_output = self.openstack(
            'router set '
            + '--distributed '
            + '--external-gateway public '
            + router_name
        )
        self.assertOutput('', cmd_output)

        cmd_output = self.openstack(
            'router show ' + router_name,
            parse_output=True,
        )
        self.assertTrue(cmd_output["distributed"])
        self.assertIsNotNone(cmd_output["external_gateway_info"])

    def test_router_add_remove_route(self):
        network_name = uuid.uuid4().hex
        subnet_name = uuid.uuid4().hex
        router_name = uuid.uuid4().hex

        self.openstack(f'network create {network_name}')
        self.addCleanup(self.openstack, f'network delete {network_name}')

        self.openstack(
            f'subnet create {subnet_name} '
            f'--network {network_name} --subnet-range 10.0.0.0/24'
        )

        self.openstack(f'router create {router_name}')
        self.addCleanup(self.openstack, f'router delete {router_name}')

        self.openstack(f'router add subnet {router_name} {subnet_name}')
        self.addCleanup(
            self.openstack,
            f'router remove subnet {router_name} {subnet_name}',
        )

        out1 = (
            self.openstack(
                f'router add route {router_name} '
                '--route destination=10.0.10.0/24,gateway=10.0.0.10',
                parse_output=True,
            ),
        )
        self.assertEqual(1, len(out1[0]['routes']))

        self.addCleanup(self.openstack, f'router set {router_name} --no-route')

        out2 = (
            self.openstack(
                f'router add route {router_name} '
                '--route destination=10.0.10.0/24,gateway=10.0.0.10 '
                '--route destination=10.0.11.0/24,gateway=10.0.0.11',
                parse_output=True,
            ),
        )
        self.assertEqual(2, len(out2[0]['routes']))

        out3 = (
            self.openstack(
                f'router remove route {router_name} '
                '--route destination=10.0.11.0/24,gateway=10.0.0.11 '
                '--route destination=10.0.12.0/24,gateway=10.0.0.12',
                parse_output=True,
            ),
        )
        self.assertEqual(1, len(out3[0]['routes']))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Hypervisor action implementations"""

import json
import re

from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _


def _get_hypervisor_columns(item, client):
    column_map = {'name': 'hypervisor_hostname'}
    hidden_columns = ['location', 'servers']

    if sdk_utils.supports_microversion(client, '2.88'):
        hidden_columns.extend(
            [
                'current_workload',
                'disk_available',
                'local_disk_free',
                'local_disk_size',
                'local_disk_used',
                'memory_free',
                'memory_size',
                'memory_used',
                'running_vms',
                'vcpus_used',
                'vcpus',
            ]
        )
    else:
        column_map.update(
            {
                'disk_available': 'disk_available_least',
                'local_disk_free': 'free_disk_gb',
                'local_disk_size': 'local_gb',
                'local_disk_used': 'local_gb_used',
                'memory_free': 'free_ram_mb',
                'memory_used': 'memory_mb_used',
                'memory_size': 'memory_mb',
            }
        )

    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class ListHypervisor(command.Lister):
    _description = _("List hypervisors")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--matching',
            metavar='<hostname>',
            help=_(
                "Filter hypervisors using <hostname> substring"
                "Hypervisor Type and Host IP are not returned "
                "when using microversion 2.52 or lower"
            ),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        parser.add_argument(
            '--long',
            action='store_true',
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        list_opts = {}

        if parsed_args.matching and (parsed_args.marker or parsed_args.limit):
            msg = _('--matching is not compatible with --marker or --limit')
            raise exceptions.CommandError(msg)

        if parsed_args.marker:
            if not sdk_utils.supports_microversion(compute_client, '2.33'):
                msg = _(
                    '--os-compute-api-version 2.33 or greater is required to '
                    'support the --marker option'
                )
                raise exceptions.CommandError(msg)
            list_opts['marker'] = parsed_args.marker

        if parsed_args.limit:
            if not sdk_utils.supports_microversion(compute_client, '2.33'):
                msg = _(
                    '--os-compute-api-version 2.33 or greater is required to '
                    'support the --limit option'
                )
                raise exceptions.CommandError(msg)
            list_opts['limit'] = parsed_args.limit

        if parsed_args.matching:
            list_opts['hypervisor_hostname_pattern'] = parsed_args.matching

        column_headers: tuple[str, ...] = (
            "ID",
            "Hypervisor Hostname",
            "Hypervisor Type",
            "Host IP",
            "State",
        )
        columns: tuple[str, ...] = (
            'id',
            'name',
            'hypervisor_type',
            'host_ip',
            'state',
        )

        if parsed_args.long:
            if not sdk_utils.supports_microversion(compute_client, '2.88'):
                column_headers += (
                    'vCPUs Used',
                    'vCPUs',
                    'Memory MB Used',
                    'Memory MB',
                )
                columns += (
                    'vcpus_used',
                    'vcpus',
                    'memory_used',
                    'memory_size',
                )

        data = compute_client.hypervisors(**list_opts, details=True)

        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class ShowHypervisor(command.ShowOne):
    _description = _("Display hypervisor details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "hypervisor",
            metavar="<hypervisor>",
            help=_("Hypervisor to display (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        hypervisor_id = compute_client.find_hypervisor(
            parsed_args.hypervisor, ignore_missing=False, details=False
        ).id
        hypervisor = compute_client.get_hypervisor(hypervisor_id).copy()

        # Some of the properties in the hypervisor object need to be processed
        # before they get reported to the user. We spend this section
        # extracting the relevant details to be reported by modifying our
        # copy of the hypervisor object.
        aggregates = compute_client.aggregates()
        hypervisor['aggregates'] = list()
        service_details = hypervisor['service_details']

        if aggregates:
            # Hypervisors in nova cells are prefixed by "<cell>@"
            if "@" in service_details['host']:
                cell, service_host = service_details['host'].split('@', 1)
            else:
                cell = None
                service_host = service_details['host']

            if cell:
                # The host aggregates are also prefixed by "<cell>@"
                member_of = [
                    aggregate.name
                    for aggregate in aggregates
                    if cell in aggregate.name
                    and service_host in aggregate.hosts
                ]
            else:
                member_of = [
                    aggregate.name
                    for aggregate in aggregates
                    if service_host in aggregate.hosts
                ]
            hypervisor['aggregates'] = member_of

        try:
            if sdk_utils.supports_microversion(compute_client, '2.88'):
                uptime = hypervisor['uptime'] or ''
                del hypervisor['uptime']
            else:
                del hypervisor['uptime']
                uptime = compute_client.get_hypervisor_uptime(
                    hypervisor['id']
                )['uptime']
            # Extract data from uptime value
            # format: 0 up 0,  0 users,  load average: 0, 0, 0
            # example: 17:37:14 up  2:33,  3 users,
            #          load average: 0.33, 0.36, 0.34
            m = re.match(
                r"\s*(.+)\sup\s+(.+),\s+(.+)\susers?,\s+load average:\s(.+)",
                uptime,
            )
            if m:
                hypervisor['host_time'] = m.group(1)
                hypervisor['uptime'] = m.group(2)
                hypervisor['users'] = m.group(3)
                hypervisor['load_average'] = m.group(4)
        except sdk_exceptions.HttpException as exc:
            if exc.status_code != 501:
                raise

        hypervisor['service_id'] = service_details['id']
        hypervisor['service_host'] = service_details['host']
        del hypervisor['service_details']

        if not sdk_utils.supports_microversion(compute_client, '2.28'):
            # microversion 2.28 transformed this to a JSON blob rather than a
            # string; on earlier fields, do this manually
            hypervisor['cpu_info'] = json.loads(hypervisor['cpu_info'] or '{}')
        display_columns, columns = _get_hypervisor_columns(
            hypervisor, compute_client
        )
        data = utils.get_dict_properties(
            hypervisor,
            columns,
            formatters={
                'cpu_info': format_columns.DictColumn,
            },
        )

        return display_columns, data



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Group action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _format_group(group):
    columns = (
        'description',
        'domain_id',
        'id',
        'name',
    )
    column_headers = (
        'description',
        'domain_id',
        'id',
        'name',
    )
    return (
        column_headers,
        utils.get_item_properties(group, columns),
    )


class AddUserToGroup(command.Command):
    _description = _("Add user to group")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Group to contain <user> (name or ID)'),
        )
        parser.add_argument(
            'user',
            metavar='<user>',
            nargs='+',
            help=_(
                'User(s) to add to <group> (name or ID) '
                '(repeat option to add multiple users)'
            ),
        )
        common.add_group_domain_option_to_parser(parser)
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        group_id = common.find_group_id_sdk(
            identity_client, parsed_args.group, parsed_args.group_domain
        )

        result = 0
        for i in parsed_args.user:
            try:
                user_id = common.find_user_id_sdk(
                    identity_client, i, parsed_args.user_domain
                )
                identity_client.add_user_to_group(user_id, group_id)
            except Exception as e:
                result += 1
                msg = _("%(user)s not added to group %(group)s: %(e)s") % {
                    'user': i,
                    'group': parsed_args.group,
                    'e': e,
                }
                LOG.error(msg)
        if result > 0:
            total = len(parsed_args.user)
            msg = (
                _(
                    "%(result)s of %(total)s users not added to group "
                    "%(group)s."
                )
            ) % {
                'result': result,
                'total': total,
                'group': parsed_args.group,
            }
            raise exceptions.CommandError(msg)


class CheckUserInGroup(command.Command):
    _description = _("Check user membership in group")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Group to check (name or ID)'),
        )
        parser.add_argument(
            'user',
            metavar='<user>',
            help=_('User to check (name or ID)'),
        )
        common.add_group_domain_option_to_parser(parser)
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        user_id = common.find_user_id_sdk(
            identity_client,
            parsed_args.user,
            parsed_args.user_domain,
            validate_actor_existence=False,
        )
        group_id = common.find_group_id_sdk(
            identity_client,
            parsed_args.group,
            parsed_args.group_domain,
            validate_actor_existence=False,
        )

        user_in_group = False
        try:
            user_in_group = identity_client.check_user_in_group(
                user_id, group_id
            )
        except sdk_exc.ForbiddenException:
            # Assume False if forbidden
            pass
        if user_in_group:
            msg = _("%(user)s in group %(group)s\n") % {
                'user': parsed_args.user,
                'group': parsed_args.group,
            }
            self.app.stdout.write(msg)
        else:
            msg = _("%(user)s not in group %(group)s\n") % {
                'user': parsed_args.user,
                'group': parsed_args.group,
            }
            self.app.stderr.write(msg)


class CreateGroup(command.ShowOne):
    _description = _("Create new group")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<group-name>',
            help=_('New group name'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain to contain new group (name or ID)'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New group description'),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing group'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if parsed_args.domain:
            kwargs['domain_id'] = common.find_domain_id_sdk(
                identity_client, parsed_args.domain
            )

        try:
            group = identity_client.create_group(**kwargs)
        except sdk_exc.ConflictException:
            if parsed_args.or_show:
                if parsed_args.domain:
                    group = identity_client.find_group(
                        parsed_args.name,
                        domain_id=parsed_args.domain,
                        ignore_missing=False,
                    )
                else:
                    group = identity_client.find_group(
                        parsed_args.name,
                        ignore_missing=False,
                    )
                LOG.info(_('Returning existing group %s'), group.name)
            else:
                raise

        return _format_group(group)


class DeleteGroup(command.Command):
    _description = _("Delete group(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'groups',
            metavar='<group>',
            nargs="+",
            help=_('Group(s) to delete (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain containing group(s) (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        errors = 0
        for group in parsed_args.groups:
            try:
                group_id = common.find_group_id_sdk(
                    identity_client, group, parsed_args.domain
                )
                identity_client.delete_group(group_id)
            except Exception as e:
                errors += 1
                LOG.error(
                    _(
                        "Failed to delete group with "
                        "name or ID '%(group)s': %(e)s"
                    ),
                    {'group': group, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.groups)
            msg = _("%(errors)s of %(total)s groups failed to delete.") % {
                'errors': errors,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListGroup(command.Lister):
    _description = _("List groups")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Filter group list by <domain> (name or ID)'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter group list by <user> (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        domain = None
        if parsed_args.domain:
            domain = common.find_domain_id_sdk(
                identity_client, parsed_args.domain
            )

        data = []
        if parsed_args.user:
            user = common.find_user_id_sdk(
                identity_client,
                parsed_args.user,
                parsed_args.user_domain,
            )
            if domain:
                # NOTE(0weng): The API doesn't actually support filtering
                # additionally by domain_id, so this doesn't really do
                # anything.
                data = identity_client.user_groups(user, domain_id=domain)
            else:
                data = identity_client.user_groups(user)
        else:
            if domain:
                data = identity_client.groups(domain_id=domain)
            else:
                data = identity_client.groups()

        # List groups
        columns: tuple[str, ...] = ('ID', 'Name')
        if parsed_args.long:
            columns += ('Domain ID', 'Description')

        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class RemoveUserFromGroup(command.Command):
    _description = _("Remove user from group")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Group containing <user> (name or ID)'),
        )
        parser.add_argument(
            'user',
            metavar='<user>',
            nargs='+',
            help=_(
                'User(s) to remove from <group> (name or ID) '
                '(repeat option to remove multiple users)'
            ),
        )
        common.add_group_domain_option_to_parser(parser)
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        group_id = common.find_group_id_sdk(
            identity_client, parsed_args.group, parsed_args.group_domain
        )

        result = 0
        for i in parsed_args.user:
            try:
                user_id = common.find_user_id_sdk(
                    identity_client, i, parsed_args.user_domain
                )
                identity_client.remove_user_from_group(user_id, group_id)
            except Exception as e:
                result += 1
                msg = _("%(user)s not removed from group %(group)s: %(e)s") % {
                    'user': i,
                    'group': parsed_args.group,
                    'e': e,
                }
                LOG.error(msg)
        if result > 0:
            total = len(parsed_args.user)
            msg = (
                _(
                    "%(result)s of %(total)s users not removed from group "
                    "%(group)s."
                )
            ) % {
                'result': result,
                'total': total,
                'group': parsed_args.group,
            }
            raise exceptions.CommandError(msg)


class SetGroup(command.Command):
    _description = _("Set group properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Group to modify (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain containing <group> (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New group name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New group description'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        group = common.find_group_id_sdk(
            identity_client, parsed_args.group, parsed_args.domain
        )
        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        identity_client.update_group(group, **kwargs)


class ShowGroup(command.ShowOne):
    _description = _("Display group details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_('Group to display (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain containing <group> (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        if parsed_args.domain:
            domain = common.find_domain_id_sdk(
                identity_client, parsed_args.domain
            )
            group = identity_client.find_group(
                parsed_args.group, domain_id=domain, ignore_missing=False
            )
        else:
            group = identity_client.find_group(
                parsed_args.group, ignore_missing=False
            )

        return _format_group(group)



#   Copyright 2021 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from osc_lib import exceptions

from openstackclient.network.v2 import local_ip_association
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes_v2
from openstackclient.tests.unit.network.v2 import fakes as network_fakes


class TestLocalIPAssociation(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()
        self.local_ip = network_fakes.create_one_local_ip()
        self.fixed_port = network_fakes.create_one_port()
        self.project = identity_fakes_v2.FakeProject.create_one_project()
        self.network_client.find_port.return_value = self.fixed_port


class TestCreateLocalIPAssociation(TestLocalIPAssociation):
    def setUp(self):
        super().setUp()
        self.new_local_ip_association = (
            network_fakes.create_one_local_ip_association(
                attrs={
                    'fixed_port_id': self.fixed_port.id,
                    'local_ip_id': self.local_ip.id,
                }
            )
        )
        self.network_client.create_local_ip_association.return_value = (
            self.new_local_ip_association
        )

        self.network_client.find_local_ip.return_value = self.local_ip

        # Get the command object to test
        self.cmd = local_ip_association.CreateLocalIPAssociation(
            self.app, None
        )

        self.columns = (
            'local_ip_address',
            'fixed_port_id',
            'fixed_ip',
            'host',
        )

        self.data = (
            self.new_local_ip_association.local_ip_address,
            self.new_local_ip_association.fixed_port_id,
            self.new_local_ip_association.fixed_ip,
            self.new_local_ip_association.host,
        )

    def test_create_no_options(self):
        arglist = [
            self.new_local_ip_association.local_ip_id,
            self.new_local_ip_association.fixed_port_id,
        ]
        verifylist = [
            ('local_ip', self.new_local_ip_association.local_ip_id),
            ('fixed_port', self.new_local_ip_association.fixed_port_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_local_ip_association.assert_called_once_with(
            self.new_local_ip_association.local_ip_id,
            **{
                'fixed_port_id': self.new_local_ip_association.fixed_port_id,
            },
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(data))

    def test_create_all_options(self):
        arglist = [
            self.new_local_ip_association.local_ip_id,
            self.new_local_ip_association.fixed_port_id,
            '--fixed-ip',
            self.new_local_ip_association.fixed_ip,
        ]
        verifylist = [
            ('local_ip', self.new_local_ip_association.local_ip_id),
            ('fixed_port', self.new_local_ip_association.fixed_port_id),
            ('fixed_ip', self.new_local_ip_association.fixed_ip),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_local_ip_association.assert_called_once_with(
            self.new_local_ip_association.local_ip_id,
            **{
                'fixed_port_id': self.new_local_ip_association.fixed_port_id,
                'fixed_ip': self.new_local_ip_association.fixed_ip,
            },
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(data))


class TestDeleteLocalIPAssociation(TestLocalIPAssociation):
    def setUp(self):
        super().setUp()
        self._local_ip_association = (
            network_fakes.create_local_ip_associations(
                count=2,
                attrs={
                    'local_ip_id': self.local_ip.id,
                },
            )
        )
        self.network_client.delete_local_ip_association.return_value = None

        self.network_client.find_local_ip.return_value = self.local_ip

        # Get the command object to test
        self.cmd = local_ip_association.DeleteLocalIPAssociation(
            self.app, None
        )

    def test_local_ip_association_delete(self):
        arglist = [
            self.local_ip.id,
            self._local_ip_association[0].fixed_port_id,
        ]
        verifylist = [
            ('local_ip', self.local_ip.id),
            ('fixed_port_id', [self._local_ip_association[0].fixed_port_id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_local_ip_association.assert_called_once_with(
            self.local_ip.id,
            self._local_ip_association[0].fixed_port_id,
            ignore_missing=False,
        )

        self.assertIsNone(result)

    def test_multi_local_ip_associations_delete(self):
        arglist = []
        fixed_port_id = []

        arglist.append(str(self.local_ip))

        for a in self._local_ip_association:
            arglist.append(a.fixed_port_id)
            fixed_port_id.append(a.fixed_port_id)

        verifylist = [
            ('local_ip', str(self.local_ip)),
            ('fixed_port_id', fixed_port_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for a in self._local_ip_association:
            calls.append(
                call(a.local_ip_id, a.fixed_port_id, ignore_missing=False)
            )

        self.network_client.delete_local_ip_association.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_local_ip_association_delete_with_exception(self):
        arglist = [
            self.local_ip.id,
            self._local_ip_association[0].fixed_port_id,
            'unexist_fixed_port_id',
        ]
        verifylist = [
            ('local_ip', self.local_ip.id),
            (
                'fixed_port_id',
                [
                    self._local_ip_association[0].fixed_port_id,
                    'unexist_fixed_port_id',
                ],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        delete_mock_result = [None, exceptions.CommandError]

        self.network_client.delete_local_ip_association = mock.MagicMock(
            side_effect=delete_mock_result
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 Local IP Associations failed to delete.', str(e)
            )

        self.network_client.delete_local_ip_association.assert_any_call(
            self.local_ip.id, 'unexist_fixed_port_id', ignore_missing=False
        )
        self.network_client.delete_local_ip_association.assert_any_call(
            self.local_ip.id,
            self._local_ip_association[0].fixed_port_id,
            ignore_missing=False,
        )


class TestListLocalIPAssociation(TestLocalIPAssociation):
    columns = (
        'Local IP ID',
        'Local IP Address',
        'Fixed port ID',
        'Fixed IP',
        'Host',
    )

    def setUp(self):
        super().setUp()
        self.local_ip_associations = (
            network_fakes.create_local_ip_associations(
                count=3,
                attrs={
                    'local_ip_id': self.local_ip.id,
                    'fixed_port_id': self.fixed_port.id,
                },
            )
        )
        self.data = []
        for lip_assoc in self.local_ip_associations:
            self.data.append(
                (
                    lip_assoc.local_ip_id,
                    lip_assoc.local_ip_address,
                    lip_assoc.fixed_port_id,
                    lip_assoc.fixed_ip,
                    lip_assoc.host,
                )
            )
        self.network_client.local_ip_associations.return_value = (
            self.local_ip_associations
        )

        self.network_client.find_local_ip.return_value = self.local_ip

        self.network_client.find_port.return_value = self.fixed_port
        # Get the command object to test
        self.cmd = local_ip_association.ListLocalIPAssociation(self.app, None)

    def test_local_ip_association_list(self):
        arglist = [self.local_ip.id]
        verifylist = [('local_ip', self.local_ip.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.local_ip_associations.assert_called_once_with(
            self.local_ip, **{}
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(list(data)))

    def test_local_ip_association_list_all_options(self):
        arglist = [
            '--fixed-port',
            self.local_ip_associations[0].fixed_port_id,
            '--fixed-ip',
            self.local_ip_associations[0].fixed_ip,
            '--host',
            self.local_ip_associations[0].host,
            self.local_ip_associations[0].local_ip_id,
        ]

        verifylist = [
            ('fixed_port', self.local_ip_associations[0].fixed_port_id),
            ('fixed_ip', self.local_ip_associations[0].fixed_ip),
            ('host', self.local_ip_associations[0].host),
            ('local_ip', self.local_ip_associations[0].local_ip_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        attrs = {
            'fixed_port_id': self.local_ip_associations[0].fixed_port_id,
            'fixed_ip': self.local_ip_associations[0].fixed_ip,
            'host': self.local_ip_associations[0].host,
        }

        self.network_client.local_ip_associations.assert_called_once_with(
            self.local_ip, **attrs
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(list(data)))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkTrunkTests(common.NetworkTests):
    """Functional tests for Network Trunks"""

    def setUp(self):
        super().setUp()

        if not self.is_extension_enabled("trunk"):
            self.skipTest("No trunk extension present")

        network_name = uuid.uuid4().hex
        subnet_name = uuid.uuid4().hex
        self.parent_port_name = uuid.uuid4().hex
        self.sub_port_name = uuid.uuid4().hex

        self.openstack(f'network create {network_name}')
        self.addCleanup(self.openstack, f'network delete {network_name}')

        self.openstack(
            f'subnet create {subnet_name} '
            f'--network {network_name} --subnet-range 10.0.0.0/24'
        )
        self.openstack(
            f'port create {self.parent_port_name} --network {network_name}'
        )
        self.addCleanup(self.openstack, f'port delete {self.parent_port_name}')
        json_out = self.openstack(
            f'port create {self.sub_port_name} --network {network_name} -f json'
        )
        self.sub_port_id = json.loads(json_out)['id']
        self.addCleanup(self.openstack, f'port delete {self.sub_port_name}')

    def test_network_trunk_create_delete(self):
        trunk_name = uuid.uuid4().hex
        self.openstack(
            f'network trunk create {trunk_name} --parent-port {self.parent_port_name} -f json '
        )
        raw_output = self.openstack('network trunk delete ' + trunk_name)
        self.assertEqual('', raw_output)

    def test_network_trunk_list(self):
        trunk_name = uuid.uuid4().hex
        json_output = json.loads(
            self.openstack(
                f'network trunk create {trunk_name} --parent-port {self.parent_port_name} -f json '
            )
        )
        self.addCleanup(self.openstack, 'network trunk delete ' + trunk_name)
        self.assertEqual(trunk_name, json_output['name'])

        json_output = json.loads(self.openstack('network trunk list -f json'))
        self.assertIn(trunk_name, [tr['Name'] for tr in json_output])

    def test_network_trunk_set_unset(self):
        trunk_name = uuid.uuid4().hex
        json_output = json.loads(
            self.openstack(
                f'network trunk create {trunk_name} --parent-port {self.parent_port_name} -f json '
            )
        )
        self.addCleanup(self.openstack, 'network trunk delete ' + trunk_name)
        self.assertEqual(trunk_name, json_output['name'])

        self.openstack('network trunk set --enable ' + trunk_name)

        json_output = json.loads(
            self.openstack('network trunk show -f json ' + trunk_name)
        )
        self.assertTrue(json_output['is_admin_state_up'])

        # Add subport to trunk
        self.openstack(
            'network trunk set '
            + f'--subport port={self.sub_port_name},segmentation-type=vlan,segmentation-id=42 '
            + trunk_name
        )
        json_output = json.loads(
            self.openstack('network trunk show -f json ' + trunk_name)
        )
        self.assertEqual(
            [
                {
                    'port_id': self.sub_port_id,
                    'segmentation_id': 42,
                    'segmentation_type': 'vlan',
                }
            ],
            json_output['sub_ports'],
        )

        # Remove subport from trunk
        self.openstack(
            'network trunk unset '
            + trunk_name
            + ' --subport '
            + self.sub_port_name
        )
        json_output = json.loads(
            self.openstack('network trunk show -f json ' + trunk_name)
        )
        self.assertEqual([], json_output['sub_ports'])

    def test_network_trunk_list_subports(self):
        trunk_name = uuid.uuid4().hex
        json_output = json.loads(
            self.openstack(
                f'network trunk create {trunk_name} --parent-port {self.parent_port_name} '
                f'--subport port={self.sub_port_name},segmentation-type=vlan,segmentation-id=42 '
                '-f json '
            )
        )
        self.addCleanup(self.openstack, 'network trunk delete ' + trunk_name)
        self.assertEqual(trunk_name, json_output['name'])

        json_output = json.loads(
            self.openstack(
                f'network subport list --trunk {trunk_name} -f json'
            )
        )
        self.assertEqual(
            [
                {
                    'Port': self.sub_port_id,
                    'Segmentation ID': 42,
                    'Segmentation Type': 'vlan',
                }
            ],
            json_output,
        )
